#!/usr/bin/env python
# encoding: utf-8
"""
anchor_pipeline.py

Created by Rod Docking on 2011-07-22.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre.
All rights reserved.
"""

# Standard Library
import sys
import os
from optparse import OptionParser, OptionGroup
import shutil
import logging

# Anchor libraries
from lib.assembly_parsers import trim_short_contigs
from lib.assembly_parsers import filter_short_exact_matches, clean_fasta_headers
from lib.alignment_runner import align_all_reads, write_bam_list
from lib.cluster_job import Cluster_job
from lib.variant_runner import call_variants_and_indels
from lib.fasta_laundry import build_header_dict
from lib.anchor_snp_indel_corrector import scrub_assembly
from lib.anchor_assembly_runner import run_anchor_assemblies
from lib.cluster_opt import parse_cluster_opts

# Check Anchor pre-requisites
from lib.check_prerequisites import check_prerequisites

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] original_ref.fa READS_LIST OUTPUT_ROOT"
    parser = OptionParser(usage)
    
    group_one = OptionGroup(parser, "Filtering")
    group_one.add_option("-r", "--clean",
        action="store_true", dest="clean",
        help="Remove long fasta header comments")
    group_one.add_option("-t", "--trim",
        action="store", type="int", dest="trim",
        help="Trim contigs shorter than this threshold")
    group_one.add_option("-f", "--filter",
        action="store", type="int", dest="filter",
        help="""Filter contigs shorter than this threshold,
                that have exact matches elsewhere in the assembly""")
    parser.add_option_group(group_one)
    
    group_two = OptionGroup(parser, "Tasks")
    group_two.add_option("-S", "--scrub",
        action="store_true", dest="scrub",
        help="Scrub contigs to correct poorly-supported base calls and indels")
    group_two.add_option("-A", "--anchor",
        action="store_true", dest="anchor",
        help="Run the Anchor local assembler to fill in scaffold gaps")
    parser.add_option_group(group_two)
    
    group_three = OptionGroup(parser, "Run Modes")
    group_three.add_option("-F", "--finalize",
        action="store_true", dest="finalize",
        help="Produce final output files")
    parser.add_option_group(group_three)
    
    group_four = OptionGroup(parser, "Run Options")
    group_four.add_option("-c", "--cluster",
        action="store", dest="cluster",
        help="Run jobs on an SGE cluster instead of locally (requires configuration)")
    group_four.add_option("-b", "--bam",
        action="store_true", dest="bam",
        help="Reads are stored in BAM files")
    group_four.add_option("-e", "--min_extend",
        action="store", type='int', dest="min_extend",
        help="Minimum contig length to attempt extension [500]",
        default="500")
    group_four.add_option("-n", "--min_gap",
        action="store", type='int', dest="min_gap",
        help="Minimum gap size to consider for assembly [2]",
        default=2)
    group_four.add_option("-d", "--min_cov",
        action="store", type='int', dest="min_cov",
        help="Minimum coverage to call a consensus base [2]",
        default=2)
    group_four.add_option("-o", "--max_overlap",
        action="store", type='float', dest="max_overlap",
        help="""Maximum tier overlap to use for assembly, 
as a proportion of read-lenth [0.75]""",
        default=0.75)
    group_four.add_option("-a", "--no_align",
        action="store", type='string', dest="no_align",
        help="""File containing list of BAM files already aligned
        against input assembly""",
        default=None)
    group_four.add_option("-v", "--no_var",
        action="store", type='string', dest="no_var",
        help="""Name of VCF file for variants already inferred
        from input assembly""",
        default=None)
    parser.add_option_group(group_four)
    
    (opt, args) = parser.parse_args()

    # Check for all pre-requisites AFTER displaying options
    #  This lets people still use -h
    check_prerequisites()
    
    # Check for proper number of arguments
    if not (len(args) == 3):
        parser.error("incorrect number of arguments")
    
    # Get cluster options
    if opt.cluster:
        cluster_opt = parse_cluster_opts(opt.cluster)
        cluster = True
    else:
        cluster_opt = parse_cluster_opts('local')
        cluster = False
    
    # Require new read alignments for finalize mode
    # Disallow scrubbing and Anchor local assemblies for finalize mode
    if opt.finalize and (opt.no_align or opt.no_var):
        parser.error("New read alignments are required for finalize mode")
    if opt.finalize and (opt.scrub or opt.anchor or opt.trim or opt.filter):
        parser.error("No assembly-cleaning tasks can be performed for finalize mode")
        
    return opt, args, cluster, cluster_opt


def start_log(output_root, opt, args):
    """
    Set up the log file for this run
    """
    
    logger = logging.getLogger("anchor_pipeline")
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler("{0}.log".format(output_root), 'a')
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    
    # Log initial command
    logger.info("---------------------------")
    logger.info("START: {0}".format(sys.argv))
    logger.info("---------------------------")
    
    # Log options and arguments
    logger.debug("Options: {0}".format(opt))
    logger.debug("Arguments: {0}".format(args))
    
    return logger

def main():
    """
    Main pipeline control flow script for the Anchor pipeline
    """
    
    # Parse command-line options
    opt, args, cluster, cluster_opt = parse_options()
    current_assembly, reads_list, output_root = args

    # Start log-file
    logger = start_log(output_root, opt, args)
    
    # Remove header comments from input Fasta file
    if opt.clean:
        current_assembly = clean_fasta_headers(output_root, current_assembly, logger)
    
    # Trim short contigs from the assembly
    if opt.trim:
        current_assembly = trim_short_contigs(output_root, current_assembly, opt.trim, logger)
    
    # Filter exactly-matching short contigs from the assembly
    if opt.filter:
        current_assembly = filter_short_exact_matches(output_root, current_assembly,
                                                      opt.filter, logger)
    
    # Run final alignments and variant calls if necessary
    if opt.finalize:

      # Copy current assembly to final name
      final_assembly = "{0}.final.fa".format(output_root)
      shutil.copy(current_assembly, final_assembly)
      current_assembly = final_assembly
    
    # Align reads to the assembly
    if opt.no_align:
        logger.debug("Using alignments from {0}".format(opt.no_align))
        bam_list = open(opt.no_align, 'r').readlines()
        bam_list = [b.rstrip() for b in bam_list]
    else:
        bam_list = align_all_reads(current_assembly, reads_list, output_root,
                                    logger, bam=opt.bam,
                                    cluster=cluster, cluster_opt=cluster_opt)
    
    # Dump the list of aligned bam files to a text file
    bam_list_file = write_bam_list(bam_list, output_root)
        
    # Call variants and short indels
    if opt.no_var:
        logger.debug("Using variant calls from {0}".format(opt.no_var))
        variant_calls = opt.no_var
    else:
        variant_calls = call_variants_and_indels(current_assembly, bam_list_file,
                                                output_root, logger, cluster=cluster,
                                                cluster_opt=cluster_opt)
    
    # Scrub the assembly to remove small errors
    if opt.scrub:
        scrubbed_assembly = "{0}.scrub.fa".format(output_root)
        scrub_assembly(variant_calls, current_assembly, scrubbed_assembly,
                      logger)
        current_assembly = scrubbed_assembly
    
    # Run the Anchor local assembler to fill in small gaps and extend contigs
    if opt.anchor:
        anchored_assembly, novel_seq_file = run_anchor_assemblies(current_assembly,
                                             bam_list_file, output_root, logger,
                                             opt.min_extend, opt.max_overlap, opt.min_cov, 
                                             opt.min_gap, cluster=cluster, cluster_opt=cluster_opt)
        current_assembly = anchored_assembly
        
    # Summarize results and clean up old files
    logger.info("---------------------------")
    logger.info("FINISH")
    logger.info("---------------------------")


if __name__ == '__main__':
    main()
    