#!/usr/bin/env python
# encoding: utf-8
"""
anchor_setup.py

Created by Rod Docking on 2011-08-22.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), ".."))
from cluster_job import Cluster_job, submit_cluster_or_local, hold_for_job_completion, remove_submission_files

def extract_pairs(read_files):
    """
    From a list of read files, extract pairs.  Assumes that if pairs are present, that
        they are sequentially listed.  The intent is to remove singleton files
    """

    # Set up an iterator
    i = iter(read_files)
    
    # Match so we can compare everything except the file number
    file_pattern = re.compile("^(.+)(_[123]_[eqc])(.+)$")
    
    # Trys to match things like:
    #  /genesis/scratch/analysis/analysis5/HS1600/61FE1AAXX_7/61FE1AAXX_7_2_export.txt.bz2
    #  /genesis/scratch/analysis/analysis5/HS1600/61U0NAAXX_5/hg19a/61U0NAAXX_5_1_concat_qseq.txt
    #   *_1_qseq.txt
    # But NOT:
    #   /genesis/scratch/analysis/analysis5/HS1600/42TJ9AAXX_5/42TJ9AAXX_5_export.txt.bz2

    while i:

        # Check the first two files
        first = i.next()
        second = i.next()
        
        # Match out everything but the file number
        search_one = re.search(file_pattern, first)
        search_two = re.search(file_pattern, second)
        
        # If both match
        if search_one and search_two:
            groups_one = search_one.groups()
            groups_two = search_two.groups()
        
            if (groups_one[0] == groups_two[0] and  
                groups_one[2] == groups_two[2]):
                yield first, second
        
        # If just the first matches
        elif search_one:
            third = i.next()
            search_three = re.search(file_pattern, third)
            
            if search_three:
                groups_three = search_three.groups()
                if (groups_one[0] == groups_three[0] and  
                    groups_one[2] == groups_three[2]):
                    yield first, third
        
        # If just the second matches    
        elif search_two:
            third = i.next()
            search_three = re.search(file_pattern, third)
            if search_three:
                groups_three = search_three.groups()
                if (groups_two[0] == groups_three[0] and  
                    groups_two[2] == groups_three[2]):
                    yield second, third

def main():
    """
    Setup for an Anchor run
    
    Usage: python anchor_setup.py assembly-contigs.fa read_list anchor_dir read_storage_dir
    """
    
    # Get full paths for all args at this point
    contigs_file, read_list, anchor_dir, read_dir = [os.path.abspath(arg) for arg in sys.argv[1:]]
    
    print "Contigs: ", contigs_file
    print "Read list:", read_list
    print "Anchor dir:", anchor_dir
    print "Read Dir:", read_dir
    
    # Set up Anchor directory and reads directory
    # NOTE - these are set to FAIL if the directory already exists
    # We don't want to accidentally over-write or re-run Anchor
    os.mkdir(anchor_dir)
    os.mkdir(read_dir)
    
    # Link in contigs file
    pathed_contigs_file = os.path.join(anchor_dir, os.path.basename(contigs_file))
    os.symlink(contigs_file, pathed_contigs_file)
    
    # For each read file, submit an abyss-tofastq job
    job_list = []
    
    # As well, re-write the input file in the Anchor directory
    new_read_list_handle = open(os.path.join(anchor_dir, 
                                os.path.basename(read_list)),'w')
    
    # Check for pairing between read files
    # ABySS can handle a mix of single-end and paired-end data, Anchor can't
    read_files = [f.rstrip() for f in open(read_list).readlines()]
    paired_read_list = []
    for pair in extract_pairs(read_files):
        paired_read_list.extend(pair)
    
    # Warn if we've done some trimming
    if not len(read_files) == len(paired_read_list):
        print "Warning!  Some files were not inferred to have proper pairs and were excluded!"
    
    for read_file in paired_read_list:
        
        # Rename and set up path
        renamed_read_file = os.path.basename(read_file) + '.fq.gz'
        pathed_read_file = os.path.join(read_dir, renamed_read_file)
        new_read_list_handle.write("{0}\n".format(pathed_read_file))
    
        # Each conversion job is a cluster job
        C = Cluster_job(jobname='abyss-to-fastq_{0}.sh'.format(
                        renamed_read_file),
                        directory=read_dir,
                        touch=True, 
                        mem_free='4G', mem_token='4G',
                        queue='all.q',)
        
        # Run abyss-tofastq and gzip
        C.append("abyss-tofastq {0} | gzip -c > {1}".format(read_file, pathed_read_file))
    
        # Submit cluster job and append ID to job list
        job_list = submit_cluster_or_local(job_list, C, True)
    
    new_read_list_handle.close()
    

if __name__ == '__main__':
    main()

