#!/usr/bin/env python
# encoding: utf-8
"""
vcf_record.py

Created by Rod Docking on 2011-02-10.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import subprocess

class VCF_record:
    """Class for encapsulating a single line of VCF output from samtools mpileup and bcftools"""
    
    def __init__(self, line):
        """Parse a line of samtools-style VCF output"""
        
        line = line.rstrip()
        
        # Skip header comments
        if line[0] == '#':
            self.header = True
            return
        
        # Parse the VCF input
        c = line.split('\t')
        
        self.header = False
        self.raw = line
        self.chrom = c[0]
        self.pos = int(c[1])
        self.id = c[2]
        self.ref = c[3]
        self.alt = c[4]
        self.qual = c[5]
        self.filter = c[6]
        
        # Split the 'INFO' field into a list
        self.info = c[7].split(';')
        
        # Leaving info for now, but also making an info_dict
        self.info_dict = {}
        for i in self.info:
            d = i.split('=')
            if len(d) == 2:
                self.info_dict[d[0]] = d[1]
            else:
                self.info_dict[d[0]] = ""
        
        # Split the genotype information into a dictionary
        format_keys = c[8].split(':')
        format_vals = c[9].split(':')
        self.genotype = {}
        
        for k,v in zip(format_keys, format_vals):
            self.genotype[k] = v
    
    # Utility functions (mainly to match the equivalent ones in pileup_record)
    def is_SNP(self):
        """Is this a SNP call?"""
        return False if (self.info[0] == 'INDEL') else True
    
    def is_indel(self):
        """Is this an indel call?"""
        return True if (self.info[0] == 'INDEL') else False
    
    def is_homozygous(self):
        """Is this a homozygous SNP?"""
        return True if (self.genotype['GT'] == '1/1') else False
    
    def is_heterozygous(self):
        """is this a heterozygous SNP?"""
        return True if (self.genotype['GT'] == '0/1') else False
    
    def print_VCF(self):
        """Return VCF-formatted output"""
        
        string_list = [str(c) for c in [self.chrom, self.pos, self.id, self.ref, self.alt, self.qual]]
        
        vcf_string_one = "\t".join(string_list)
        vcf_string_two = ";".join(self.info)
        
        if self.genotype.has_key('GT'):
            vcf_string_three = "GT:PL:GQ\t{0}:{1}:{2}".format(
                self.genotype['GT'], self.genotype['PL'], self.genotype['GQ'])
        else:
            vcf_string_three = ""
        
        return "\t".join([vcf_string_one, vcf_string_two, vcf_string_three])

def bcftools_view(bcf_file):
    """Run bcftools view to return VCF output"""
    
    proc = subprocess.Popen(
            ['bcftools', 'view', bcf_file],
            shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
    
    return proc.stdout

def main():
    """Test VCF parsing"""
    
    for line in open(sys.argv[1], 'r'):
        
        V = VCF_record(line)
        
        if V:
            
            print "---------"
            print V.raw
            print "Chrom", V.chrom
            print "ID", V.id
            print "Ref:", V.ref
            print "Alt:", V.alt
            print "Qual:", V.qual
            print "Filter", V.filter
            print "Info:", V.info
            
            print "Genotype:", V.genotype
            
            print "SNP?", V.is_SNP()
            print "Indel?", V.is_indel()
            print "Homozygous?", V.is_homozygous()
            print "Heterozygous?", V.is_heterozygous()



if __name__ == '__main__':
    main()

