#!/usr/bin/env python
# encoding: utf-8
"""
blat_validation.py

Created by Rod Docking on 2011-08-09.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re
import collections

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), ".."))
from lib.sam_record import SAM_record, bam_open

MIN_IDENTITY = 0.98

def split_alignments(alignments):
    """
    Split a set of alignments into two lists - one for each haplotype
    """

    align_dict = {
        'hap1_target' : [],
        'hap2_target' : [],
        'hap1_novel'  : [],
        'hap2_novel'  : [],
    }

    for S in alignments:
        if "target" in S.qname:
            if S.rname[-1] == "1":
                align_dict['hap1_target'].append(S)
            elif S.rname[-1] == "2":
                align_dict['hap2_target'].append(S)
        else:
            if S.rname[-1] == "1":
                align_dict['hap1_novel'].append(S)
            elif S.rname[-1] == "2":
                align_dict['hap2_novel'].append(S)

    return align_dict

def write_extension_BED(alignments, handle, result, validation):
    """
    Write extension information to the appropriate BED file
    """
    
    for S in alignments:
        
        if "target_back" in S.qname:
            start, stop = S.pos, S.pos+100
        elif "target_front" in S.qname:
            start = S.pos + len(S.seq) - 100
            stop = start + 100
        
        handle.write("{0}\t{1}\t{2}\tID={3};Note={4}_{5}\n".format(
            S.rname, start, stop, S.qname, "_".join(result.split()[-2:]), 
            validation))
    
    return

def write_gap_BED(alignments, handle, result, validation):
    """
    Write extension information to the appropriate BED file
    """

    for S in alignments:

        if "target_front" in S.qname:
            start, stop = S.pos + len(S.seq) - 100, S.pos + len(S.seq)
            handle.write("{0}\t{1}\t{2}\tID={3};Note={4}_{5}\n".format(
                S.rname, start, stop, S.qname, "_".join(result.split()[-2:]),
                validation))

    return

def check_overlap(q1, q2):
    """
    Confirm that two alignments overlap on the reference
    """
    
    # Check for same target sequence
    if not (q1.rname == q2.rname):
        return False
    
    # Check for overlap length
    q1_range, q2_range = [],[]    
    for i in range(q1.pos, q1.pos + len(q1.seq)):
        q1_range.append(i)
    for i in range(q2.pos, q2.pos + len(q2.seq)):
        q2_range.append(i)
    
    q1_multiset = collections.Counter(q1_range)
    q2_multiset = collections.Counter(q2_range)
    overlap = list((q1_multiset & q2_multiset).elements())
    print "Overlap length", len(overlap), q1.qname, q2.qname
    
    if not len(overlap) >= 150:
        return False
        
    return True        

def validate_extension(result, align_dict, hap1_handle, hap2_handle, result_dict):
    """
    Validate an extension assembly    
    """

    # Test whether Anchor-marked successes are truly successes
    if "extend" in result:
        extension_success_validated = False
                    
        # At least one haplotype target should exactly match the reference
        match_list_targets = [S.identity >= MIN_IDENTITY for 
                               S in align_dict['hap1_target'] + align_dict['hap2_target']]
        
        # At least one haplotype extension sequence should exactly
        #   match the reference
        match_list_novel = [((S.identity >= MIN_IDENTITY) and (not 'S' in S.cigar)) for 
                            S in align_dict['hap1_novel'] + align_dict['hap2_novel']]
        
        print "Target matches:", match_list_targets
        print "Novel matches:", match_list_novel
        
        # At least one extension should overlap with its target on the reference
        coordinate_overlap_hap1, coordinate_overlap_hap2 = False, False
        if (len(align_dict['hap1_target']) >= 1) and (len(align_dict['hap1_novel']) >= 1):
            coordinate_overlap_hap1 = check_overlap(align_dict['hap1_target'][0],
                                                    align_dict['hap1_novel'][0])

        if (len(align_dict['hap2_target']) >= 1) and (len(align_dict['hap2_novel']) >= 1):
            coordinate_overlap_hap2 = check_overlap(align_dict['hap2_target'][0],
                                                    align_dict['hap2_novel'][0])
    
        # If this is true for at least one of the haplotypes, pass
        if ((True in match_list_targets) and (True in match_list_novel) and 
            (coordinate_overlap_hap1 or coordinate_overlap_hap2)):
            extension_success_validated = True
        
        if extension_success_validated:
            result_dict['extension_success_validated'] += 1
            print "==> Validated"
            validation = "validated"
        else:
            result_dict['extension_success_failed'] += 1
            print "==> Failed"
            
            if not (True in match_list_targets):
                result_dict['extension_success_failed_bad_targets'] += 1
                print "==> extension_success_failed_bad_targets"
                validation = "fail_bad_targets"
            elif not (True in match_list_novel):
                result_dict['extension_success_failed_bad_assembly'] += 1
                print "==> extension_success_failed_bad_assembly"
                validation = "fail_bad_assembly"
            elif not (coordinate_overlap_hap1 or coordinate_overlap_hap2):
                result_dict['extension_success_failed_no_overlap'] += 1
                print "==> extension_success_failed_no_overlap"
                validation = "fail_no_overlap"
    
    # Test what happened for Anchor-marked failures 
    else:
        
        # At least one haplotype target should match the reference at >= 99% identity
        match_list_targets = [S.identity >= MIN_IDENTITY for
                            S in align_dict['hap1_target'] + align_dict['hap2_target']]
        
        if True in match_list_targets:
            result_dict['extension_fail_good_targets'] += 1
            print "==> extension_fail_good_targets"
            validation = "validated"
        else:
            result_dict['extension_fail_bad_targets'] += 1
            print "==> extension_fail_bad_targets"
            validation = "fail"

    # Mark the appropriate BED file with the end of the target
    write_extension_BED(align_dict['hap1_target'], hap1_handle, result, validation)
    write_extension_BED(align_dict['hap2_target'], hap2_handle, result, validation)
    
    return result_dict

def validate_gap_fail(align_dict, result_dict):
    """
    Validate a gap failure
    """
    
    # The front targets should have at least one good hit
    front_targets = [S for S in (align_dict['hap1_target'] + align_dict['hap2_target']) 
                    if "front" in S.qname]
    match_list_front = [S.identity >= MIN_IDENTITY for S in front_targets]
    
    # The back targets should have at least one good hit
    back_targets = [S for S in (align_dict['hap1_target'] + align_dict['hap2_target']) 
                    if "back" in S.qname]
    match_list_back = [S.identity >= MIN_IDENTITY for S in back_targets]

    if (True in match_list_front) and (True in match_list_back):
        result_dict['gap_fail_good_targets'] += 1
        print "==> Validated"
        validation = "validated"
    else:
        result_dict['gap_fail_bad_targets'] += 1
        print "==> gap_fail_bad_targets"
        validation = "fail_poor_targets"

    return result_dict, validation

def validate_gap_close(align_dict, result_dict):
    """
    Validate a gap closure
    """

    # The front targets should have at least one good hit
    front_targets = [S for S in (align_dict['hap1_target'] + align_dict['hap2_target']) 
                    if "front" in S.qname]
    match_list_front = [S.identity >= MIN_IDENTITY for S in front_targets]
    
    # The back targets should have at least one good hit
    back_targets = [S for S in (align_dict['hap1_target'] + align_dict['hap2_target']) 
                    if "back" in S.qname]
    match_list_back = [S.identity >= MIN_IDENTITY for S in back_targets]

    # The novel sequence should exactly match the reference on at least one haplotype
    novel_seqs = [S for S in (align_dict['hap1_novel'] + align_dict['hap2_novel'])]
    match_list_novel = [((S.identity >= MIN_IDENTITY) and (not 'S' in S.cigar)) for S in novel_seqs]

    # The novel sequence should overlap both of its targets on the reference
    hap1_overlap, hap2_overlap = False, False
    overlap_hap1, overlap_hap2 = [], []
    if len(align_dict['hap1_target'] + align_dict['hap1_novel']) >= 3:
        overlap_hap1 = [check_overlap(align_dict['hap1_novel'][0], a) for a in align_dict['hap1_target']]
        if (not False in overlap_hap1):
            hap1_overlap = True

    if len(align_dict['hap2_target'] + align_dict['hap2_novel']) >= 3:
        overlap_hap2 = [check_overlap(align_dict['hap2_novel'][0], a) for a in align_dict['hap2_target']]
        if (not False in overlap_hap2):
            hap2_overlap = True

    if ((True in match_list_front) and (True in match_list_back) and (True in match_list_novel) 
        and (hap1_overlap or hap2_overlap)):
        result_dict['gap_close_validated'] += 1
        print "==> Validated"
        validation = "validated"

    else:
        result_dict['gap_close_failed'] += 1
        print "==> Failed"
        
        if not ((True in match_list_front) and (True in match_list_back)):
            print "==> gap_close_failed_bad_targets"
            result_dict['gap_close_failed_bad_targets'] += 1
            validation = "fail_poor_targets"
        elif not (True in match_list_novel):
            print "==> gap_close_failed_bad_assembly"
            result_dict['gap_close_failed_bad_assembly'] += 1
            validation = "fail_poor_assembly"
        elif not (hap1_overlap or hap2_overlap):     
            print "==> gap_close_failed_no_overlap"
            result_dict['gap_close_failed_no_overlap'] += 1
            validation = "fail_poor_overlap"
            
    return result_dict, validation

def validate_gap_reduction(align_dict, result_dict):
    """
    Validate a gap closure
    """
    
    # The front targets should have at least one good hit
    front_targets = [S for S in (align_dict['hap1_target'] + align_dict['hap2_target']) 
                    if "front" in S.qname]
    match_list_front = [S.identity >= MIN_IDENTITY for S in front_targets]
    
    # The back targets should have at least one good hit
    back_targets = [S for S in (align_dict['hap1_target'] + align_dict['hap2_target']) 
                    if "back" in S.qname]
    match_list_back = [S.identity >= MIN_IDENTITY for S in back_targets]

    # The novel sequences should exactly match the reference on at least one haplotype
    # Need to split by each assembly fragment here...
    novel_seqs = [S for S in (align_dict['hap1_novel'] + align_dict['hap2_novel'])]
    novel_dict = {}
    for S in novel_seqs:
        if S.qname in novel_dict:
            novel_dict[S.qname].append(S)
        else:
            novel_dict[S.qname] = [S]

    # For each assembled sub-sequence, there should be at least one complete hit
    # Additionally, each assembled sub-sequence should overlap with at least one target
    match_list_novel, overlap_list_novel = [], []
    for novel_seq in novel_dict:
        target_match = [((S.identity >= MIN_IDENTITY) and (not 'S' in S.cigar)) for S in novel_dict[novel_seq]]
        if True in target_match:
            match_list_novel.append(True)
        else:
            match_list_novel.append(False)        
        
        overlaps = []
        for target in (front_targets + back_targets):
            overlaps.extend([check_overlap(S, target) for S in novel_dict[novel_seq]])
        if True in overlaps:
            overlap_list_novel.append(True) 
        else:
            overlap_list_novel.append(False)
    
    print match_list_front
    print match_list_back
    print match_list_novel
    print overlap_list_novel
    
    # Does this need to be more sensitive?
    if ((True in match_list_front) and (True in match_list_back) and 
         (not False in match_list_novel) and (not False in overlap_list_novel)):
        result_dict['gap_reduce_validated'] += 1
        print "==> Validated"
        validation = "validated"
    else:
        result_dict['gap_reduce_failed'] += 1
        print "==> Failed"

        if not ((True in match_list_front) and (True in match_list_back)):
            print "==> gap_reduce_fail_poor_targets"
            validation = "fail_poor_targets"
            result_dict['gap_reduce_failed_poor_targets'] += 1
        elif (False in match_list_novel):
            print "==> gap_reduce_failed_poor_assembly"
            validation = "fail_poor_assembly"
            result_dict['gap_reduce_failed_poor_assembly'] += 1
        elif (False in overlap_list_novel):
            print "==> gap_reduce_failed_poor_overlap"
            validation = "fail_poor_overlap"
            result_dict['gap_reduce_failed_poor_overlap'] += 1

    return result_dict, validation

def parse_ref_aligns(sam_dict, bam_file):
    """
    Add SAM hits to a dictionary
    """

    assembly_pattern = re.compile("^(\d+_\d{1,3})")
    
    for S in bam_open(bam_file):
        
        assembly_match = re.match(assembly_pattern, S.qname)
        assembly = assembly_match.groups()[0]
         
        if assembly in sam_dict:
            sam_dict[assembly].append(S)
        else:
            sam_dict[assembly] = [S]    
    
    return sam_dict

def calculate_gap_size(align_dict):
    """
    From a set of SAM alignments, calculate the gap size
     between the front and back targets
    """

    hap1_front_start, hap1_front_stop = 0, 0
    hap1_back_start, hap1_back_stop = 0, 0
    hap2_front_start, hap2_front_stop = 0, 0
    hap2_back_start, hap2_back_stop = 0, 0
    hap1_estimate, hap2_estimate = 0, 0
    hap1_front_rname, hap2_front_rname, hap1_back_rname, hap2_back_rname = None, None, None, None
    
    for S in align_dict['hap1_target']:
        if "front" in S.qname:
            hap1_front_start = S.pos
            hap1_front_stop = S.pos + len(S.seq)
            hap1_front_rname = S.rname
        elif "back" in S.qname:
            hap1_back_start = S.pos
            hap1_back_stop = S.pos + len(S.seq)
            hap1_back_rname = S.rname

    for S in align_dict['hap2_target']:
        if "front" in S.qname:
            hap2_front_start = S.pos
            hap2_front_stop = S.pos + len(S.seq)
            hap2_front_rname = S.rname
        elif "back" in S.qname:
            hap2_back_start = S.pos
            hap2_back_stop = S.pos + len(S.seq)
            hap2_back_rname = S.rname

    if (hap1_front_start and hap1_front_stop and hap1_back_start and hap1_back_stop):
        hap1_estimates = [(hap1_back_start - hap1_front_stop), (hap1_front_start - hap1_back_stop)]
        hap1_estimate = hap1_estimates[0] if (abs(hap1_estimates[0]) < abs(hap1_estimates[1])) else hap1_estimates[1]

    if (hap2_front_start and hap2_front_stop and hap2_back_start and hap2_back_stop):
        hap2_estimates = [(hap2_back_start - hap2_front_stop), (hap2_front_start - hap2_back_stop)]
        hap2_estimate = hap2_estimates[0] if (abs(hap2_estimates[0]) < abs(hap2_estimates[1])) else hap2_estimates[1]

    return hap1_estimate, hap2_estimate, (hap1_front_rname == hap1_back_rname), (hap2_front_rname == hap2_back_rname)

def main():
    """
    Parse alignments of assembled sequences vs. a reference and report any misassemblies
    """

    # Parse the Anchor result file to get a dict of assembly results
    assemblies = {}
    with open(sys.argv[1]) as in_handle:
        for line in in_handle:
            c = line.split()
            assemblies[c[0]] = " ".join(c[1:])
    
    # Read all alignments
    sam_dict = {}
    for sam_file in sys.argv[3:]:
        sam_dict = parse_ref_aligns(sam_dict, sam_file)
    
    # BED output files
    bed_root = sys.argv[2]
    gap_hap1_bed = open("{0}_gaps.hap1.bed".format(bed_root),'w')
    gap_hap2_bed = open("{0}_gaps.hap2.bed".format(bed_root),'w')
    ext_hap1_bed = open("{0}_ext.hap1.bed".format(bed_root),'w')
    ext_hap2_bed = open("{0}_ext.hap2.bed".format(bed_root),'w')
    
    # BED headers for IGV
    ext_hap1_bed.write("#gffTags track name=ext_hap1 type=bedDetail description=\"Haplotype 1 Extensions\"\n")
    ext_hap2_bed.write("#gffTags track name=ext_hap2 type=bedDetail description=\"Haplotype 2 Extensions\"\n")
    gap_hap1_bed.write("#gffTags track name=gap_hap1 type=bedDetail description=\"Haplotype 1 Gaps\"\n")
    gap_hap2_bed.write("#gffTags track name=gap_hap2 type=bedDetail description=\"Haplotype 2 Gaps\"\n")
    
    # Result dict
    result_dict = {
        'extension_success_validated'   : 0,
        'extension_success_failed'      : 0,
        'extension_success_failed_bad_targets'  : 0,
        'extension_success_failed_bad_assembly' : 0,
        'extension_success_failed_no_overlap' : 0,
        'extension_fail_good_targets'   : 0,
        'extension_fail_bad_targets'    : 0,
        'gap_fail_good_targets'         : 0,
        'gap_fail_bad_targets'          : 0,
        'gap_close_validated'           : 0,
        'gap_close_failed'              : 0,
        'gap_close_failed_bad_targets'  : 0,
        'gap_close_failed_bad_assembly' : 0,
        'gap_close_failed_no_overlap'   : 0,
        'gap_reduce_validated'          : 0,
        'gap_reduce_failed'             : 0,
        'gap_reduce_failed_poor_targets' : 0,
        'gap_reduce_failed_poor_assembly': 0,
        'gap_reduce_failed_poor_overlap' : 0,
    }
    
    # For each local assembly, collect reference alignments and assess results
    for assembly in sorted(assemblies):
        
        # Retrieve result and print header
        result = assemblies[assembly]
            
        print "\n", assembly, result
        print "----------" 

        # Check for alignments
        if not assembly in sam_dict:
            print "Missing SAM records for ", assembly, result
            continue        
        
        # Retrieve alignments and split by type
        alignments = sam_dict[assembly]
        align_dict = split_alignments(alignments)
         
        # Print a summary of each alignment against each haplotype
        for S in align_dict['hap1_target']:
            print "{0:<30}\t{1}\t{2}\t{3:<15}\tNM:{4}\t{5:.3f}".format(S.qname, S.rname, S.pos, S.cigar, S.edit_distance, S.identity)
        for S in align_dict['hap1_novel']:
            print "{0:<30}\t{1}\t{2}\t{3:<15}\tNM:{4}\t{5:.3f}".format(S.qname, S.rname, S.pos, S.cigar, S.edit_distance, S.identity)
        print ""
        for S in align_dict['hap2_target']:
            print "{0:<30}\t{1}\t{2}\t{3:<15}\tNM:{4}\t{5:.3f}".format(S.qname, S.rname, S.pos, S.cigar, S.edit_distance, S.identity)
        for S in align_dict['hap2_novel']:
            print "{0:<30}\t{1}\t{2}\t{3:<15}\tNM:{4}\t{5:.3f}".format(S.qname, S.rname, S.pos, S.cigar, S.edit_distance, S.identity)

        # Get estimated gap size
        if "gap" in result:
                        
            # Parse out gap estimate from assembler
            gap_pattern = re.compile("\[(\d+), (\d+)\]")
            gap_search = re.search(gap_pattern, result) 
            if gap_search.groups():
                gap_estimate = int(gap_search.groups()[1]) - int(gap_search.groups()[0])
                print "\nGap estimate from assembler: ", gap_estimate

            # Calculate the gap size from the alignments
            hap1_estimate, hap2_estimate, hap1_same_ref, hap2_same_ref = calculate_gap_size(align_dict)
            print "Hap1 gap estimate: ", hap1_estimate
            print "Hap2 gap estimate: ", hap2_estimate
            
            # Calculate the minimum absolute difference as a pct of the the assembler estimate
            diff1 = abs(gap_estimate - hap1_estimate)
            diff2 = abs(gap_estimate - hap2_estimate)
            min_abs_diff = min(diff1,diff2)
            min_pct_diff = float(min_abs_diff) / float(gap_estimate) * 100
            print "Minimum absolute difference: {0} ({1:.2f}%)".format(min_abs_diff, min_pct_diff)
            
            # Flag potential misassemblies:
            if (not (hap1_same_ref or hap2_same_ref)) or (min_pct_diff > 10.0 and min_abs_diff > 100):
                print "==> Potential misassembly!"
                
        #Analyze depending on assembly result        
        if "exten" in result:
            result_dict = validate_extension(
                result, align_dict, ext_hap1_bed, ext_hap2_bed, result_dict)
        elif "gap close" in result:
            result_dict, validation = validate_gap_close(align_dict, result_dict)
        elif "gap reduced" in result:
            result_dict, validation = validate_gap_reduction(align_dict, result_dict)
            write_gap_BED(align_dict['hap1_target'], gap_hap1_bed, result, validation)
            write_gap_BED(align_dict['hap2_target'], gap_hap2_bed, result, validation)
        elif "gap fail" in result:
            result_dict, validation = validate_gap_fail(align_dict, result_dict)
        else:
            print "Missed a case!!"
            sys.exit()    
                
    print "Results: (Pass/Fail)"
    print result_dict
    print "Gaps:"
    print "  Gap closures:"
    print "    Validated:\t{0:,d}".format(result_dict['gap_close_validated'])
    print "    Failed:\t{0:,d}".format(result_dict['gap_close_failed'])
    print "      Bad targets:\t{0:,d}".format(result_dict['gap_close_failed_bad_targets'])
    print "      Bad assembly:\t{0:,d}".format(result_dict['gap_close_failed_bad_assembly'])
    print "      No overlap:\t{0:,d}".format(result_dict['gap_close_failed_no_overlap'])
    print "  Gap reductions:"
    print "    Validated:\t{0:,d}".format(result_dict['gap_reduce_validated'])
    print "    Failed:\t{0:,d}".format(result_dict['gap_reduce_failed'])
    print "      Bad targets:\t{0:,d}".format(result_dict['gap_reduce_failed_poor_targets'])
    print "      Bad assembly:\t{0:,d}".format(result_dict['gap_reduce_failed_poor_assembly'])
    print "      No overlap:\t{0:,d}".format(result_dict['gap_reduce_failed_poor_overlap'])
    print "  Gap failures:"
    print "    Validated:\t{0:,d}".format(result_dict['gap_fail_good_targets'])
    print "    Failed:\t{0:,d}".format(result_dict['gap_fail_bad_targets'])
    print "Extensions:"
    print "  Extension success:"
    print "    Validated:\t{0:,d}".format(result_dict['extension_success_validated'])
    print "    Failed:\t{0:,d}".format(result_dict['extension_success_failed'])
    print "      Bad targets:\t{0:,d}".format(result_dict['extension_success_failed_bad_targets'])
    print "      Bad assembly:\t{0:,d}".format(result_dict['extension_success_failed_bad_assembly'])
    print "      No overlap:\t{0:,d}".format(result_dict['extension_success_failed_no_overlap'])
    print "  Extension fail:"
    print "    Validated:\t{0:,d}".format(result_dict['extension_fail_good_targets'])
    print "    Failed:\t{0:,d}".format(result_dict['extension_fail_bad_targets'])

    # print "Extensions: ", extend_pass, extend_fail
    # print "Gap closures: ", gap_close_pass, gap_close_fail
    # print "Gap reductions: ", gap_reduce_pass, gap_reduce_fail                    
            
if __name__ == '__main__':
    main()

